# Copyright (c) 2024, InfinityQ Technology, Inc.

from typing import Optional, Tuple

import numpy as np

from .variable_list import VariableVectorList
from .variable import Expression, SubVariable, Term, VariableVector, VectorExpression


def objective_matrices_from_expression(
    expr: Expression,
    variables: VariableVectorList
) -> Tuple[Optional[np.ndarray], np.ndarray]:
    """
    Extracts the bias vector and weights matrix from an expression.

    This function processes an expression and a list of variable vectors to extract
    the bias vector and weights matrix. The bias vector and weights matrix are initialized
    based on the number of variables and then populated based on the terms in the expression.

    Parameters
    ----------
    expr
        The expression from which to extract the bias vector and weights matrix.
    variables
        A list of variable vectors used in the expression.
        Each variable vector's size contributes to the overall size of the bias vector and weights matrix.

    Returns
    -------
    weights matrix if any and bias vector
    """
    if isinstance(expr, (VariableVector, VectorExpression, float, int)) or expr is None:
        raise ValueError("The provided input is not an expression")

    total_size = variables.total_variable_size()
    bias = np.zeros(total_size, dtype=np.float32)
    weights = np.zeros((total_size, total_size), dtype=np.float32)

    if isinstance(expr, SubVariable):
        expr = Term(expr, None, 1)
    if isinstance(expr, Term):
        expr = Expression([expr])

    indexed_variables = variables.index_variables()
    for term in expr.terms():
        if isinstance(term, Term):
            index_v1 = indexed_variables[term.v1().parent()] + term.v1().index()
            # linear
            if term.v2() is None:
                bias[index_v1] += term.coeff()
            # quadratic
            else:
                index_v2 = indexed_variables[term.v2().parent()] + term.v2().index()
                weights[index_v1, index_v2 ] += term.coeff()
                if index_v1 != index_v2:
                    weights[index_v2 , index_v1] += term.coeff()
        else:
            pass
            # TODO use const Term for constraints

    return weights if np.any(weights) else None, bias


# Copyright (c) 2024, InfinityQ Technology, Inc.

from unittest.mock import create_autospec, patch
import numpy as np
from titanq import Model, Vtype
from titanq.tools import configure_model_from_mps_file

def test_configure_model_from_mps_file():
    model = create_autospec(Model, instance=True)

    configure_model_from_mps_file(model, 'src/tests/test_files/mps_test.mps')

    actual_variable_calls = model.add_variable_vector.call_args_list
    expected_bounds = [(0.0, 1.0), (0.0, np.inf), (0.0, np.inf)]
    for actual_call, expected_call in zip(actual_variable_calls, expected_bounds):
        actual_var_args, _ = actual_call
        assert actual_var_args[2] == Vtype.CONTINUOUS
        np.testing.assert_array_equal(actual_var_args[3], [expected_call])
    
    actual_objective_args, _ = model.set_objective_matrices.call_args
    expected_matrix = np.zeros((3, 3), dtype=np.float32)
    expected_vector = np.array([1., 2., 1.], dtype=np.float32)
    np.testing.assert_array_equal(actual_objective_args[0], expected_matrix)
    np.testing.assert_array_equal(actual_objective_args[1], expected_vector)

    actual_eq_args, _ = model.add_equality_constraint.call_args
    expected_vector = np.array([0, -1, 1], dtype=np.float32)
    np.testing.assert_array_equal(actual_eq_args[0], expected_vector)
    assert actual_eq_args[1] == 0

    inequality_calls = model.add_inequality_constraint.call_args_list
    expected_ineq_vectors = [
        (np.array([1., 1., 0.]), [2, 4]),
        (np.array([-1., 0., -1.]), [np.nan, -3])
    ]
    for actual_call, expected_call in zip(inequality_calls, expected_ineq_vectors):
        actual_ineq_args, _ = actual_call
        np.testing.assert_array_equal(actual_ineq_args[0], expected_call[0])
        np.testing.assert_array_equal(actual_ineq_args[1], expected_call[1])
